/*
Realisiert mit AT-Mega48, 128KHz interner RC-Oszillator

Ausgnge:
PB0     Taste Spalte 1
PB1     Taste Spalte 2
PB2     Taste Spalte 3

PC0     Relais Spalte 1  - Spalten-Treiber mit Schutzdioden
PC1     Relais Spalte 2
PC2     Relais Spalte 3
PC3     Relais Spalte 4
PC4     Blinkled
PC5     Triac

PD4     Relais Zeile 1
PD5     Relais Zeile 2
PD6     Relais Zeile 3
PD7     Relais Zeile 4

Eingnge:
PB3     Taste Zeile 1
PB4     Taste Zeile 2
PB5     Taste Zeile 3

PD2     IR-Detector


==========================================
T0 - 8Bit:   1s Takt / Delay-Tic
T1 - 16Bit:  
T2 - 8bit:   Sleep-Wecker 

INT0: IR-Empfnger


Relaisspulen: 
  Zeile 1: Relais A-D Ein
  Zeile 2: Relais A-D Aus
  Zeile 3: Relais E-H Ein
  Zeile 4: Relais E-H Aus

*/

#define TIMEOUT 120L    //Minuten bis zum Abschalten
#define TIMEOUTWARN 15L //Minuten vor Abschaltung blinken


#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>


typedef unsigned char byte;


//Der Einfachheit halber: viele globale Variablen

//Wir haben insgesamt 1024 Bytes SRAM - reichlich Platz, um die Variablen statisch zu halten
typedef struct {
  long int timeout_s;    
  char *pattern;
  byte tasteBisher;
  byte irEmpfangen;
  long int blinkschwelle;
  long int abschaltschwelle;
  byte toggleBisher[8];

}Tglob;

//Alle globalen Var in einer Struct deklarieren
Tglob g;


//--------------------------------------------------------
//--------------------------------------------------------

//--------------------------------------------------------
ISR(TIMER0_OVF_vect){ //488ms Interrupt
  g.timeout_s++;  
}

//--------------------------------------------------------
ISR(TIMER2_COMPA_vect){ 
    //Sleep wird aufgeweckt
    asm("nop");
}

//--------------------------------------------------------
ISR(INT0_vect){ //IR-Empfang
    g.timeout_s=0;
    g.irEmpfangen=1;
}

//--------------------------------------------------------
void nickerchen(void){
  SMCR = 1; //SE enable; Idle-Mode
  asm("SLEEP");
  //Schlft bis T0 oder T2 berluft (also ca. 10ms)
}

//--------------------------------------------------------
void delay(int dauer_ms){
//mind. 10ms,  max 500ms  
//Auflsung: ca. 10ms
int16_t startcnt;
int16_t endeTic;
int16_t delta;
int16_t n;
    //Timer 0 auslesen
    startcnt = TCNT0;

    endeTic = dauer_ms / 2;
    //Warten, bis Timer Endwert erreicht
    do {
      nickerchen();    
      n = TCNT0;
      if (n < startcnt) startcnt-=256;
      delta = n;
      delta -= startcnt;
    }while (delta < endeTic);
}

//--------------------------------------------------------

int main (void)
{
int i,n;
byte tastencode;   //Tastencode 
unsigned char *ram;

unsigned char Relaiscode[4]={1,2,4,8};


    cli();

    DDRC = 1+2+4+ 8+16+32;
    PORTC = 0; //alles AUS

    DDRD = 16+32+64+128;
    PORTD = 0; //alles AUS

    DDRB = 1+2+4;
    PORTB = 1+2+4  +  8+16+32; //Spalten High, Pull-Ups der Zeilen ein.



    //Set up the 8-bit timer 0 (2ms)
    //-----------------------------------------
    TCCR0A = 0; //Normal Mode 
    TCCR0B = 4; // Clock/256  = 500Hz
    TCNT0 = 0; //Zhlerstand=0



/*
        CS02 CS01 CS00 Description:
        000 No clock source (Timer/Counter stopped).
        001 No prescaling
        010 /8    (From prescaler)
        011 /64   (From prescaler)
        100 /256   (From prescaler)
        101 /1024  (From prescaler)
        110 Ext T1 falling edge
        111 Ext T1 rising edge
*/


    //Set up the 16-bit timer 1 
    //---------------------------------------
    TCCR1B = 0; //Stopped
            
    //Set up the 8-bit timer 2  (20Hz Generator)
    //------------------------------------------
    TCCR2A = _BV(WGM21);  //  CTC Modus
    TCCR2B =  //_BV(FOC2)
            // _BV(WGM20)  
            //_BV(COM21)
            //_BV(COM20)
            
            _BV(CS22)  // Clock Select 100 => Teiler=256 (500Hz)
            | _BV(CS21)  
            //| _BV(CS20) 
            //_BV(CS10)
            ;
/*
        CS22 CS21 CS20 Description:
        000 No clock source (Timer/Counter stopped).
        001 No prescaling
        010 /8    (From prescaler)
        011 /32   (From prescaler)
        100 /64   (From prescaler)
        101 /128  (From prescaler)
        110 /256  (From prescaler)
        111 /1024 (From prescaler)
*/

            
    OCR2A = 4; //Zhlt 0..4, sollte 100Hz Interrupt-Frequenz ergeben
              //(Zum aufwecken aus dem Sleep-Mode)
 
    //Timer Interrupt Mask
    TIMSK0 = _BV(TOIE0);  //Timer 0 Overflow
    TIMSK2 = _BV(OCIE2A);  //Timer 2 Output Compare Match
    


    //External interrupt INT0 (PD2) bei steigender Flanke
    EICRA = 3; //IRQ bei steigender Flanke;
    EIMSK = 1; //INT0 einschalten
    

    ADCSRA = 7; //ADC abschalten (A/D Wandler)
    ACSR = 128; //ACD abschalten (Komparator)

    //Globale Variablen initialisieren
//    memset(&g, 0, sizeof(Tglob));               Da rastet der Chip aus
    ram=(unsigned char *)&g;
    for (i=0; i< sizeof(Tglob); ram[i++]=0);   //Mach ich's halt von Hand

    g.blinkschwelle = (TIMEOUT-TIMEOUTWARN)*1000/488*60;
    g.abschaltschwelle = TIMEOUT*1000/488*60;

    sei(); //Interrupts EIN. Jetzt geht's los!

    //Blinken beim Start
    for (i=0; i<10; i++){
        //LED einschalten (PC4)
        PORTC = 16;
        nickerchen(); 
        delay(50);

        //LED ausschalten (PC4)
        PORTC = 0;
        nickerchen(); 
        delay(50);
    }


    while(1){

        if (g.irEmpfangen){
            g.irEmpfangen=0;
            PORTC |= 0x10;  //Blinkled ein
            delay(20);
        }else{

            //Timeout fast erreicht?
            if (g.timeout_s > g.blinkschwelle){
                if (g.timeout_s & 1){
                    PORTC |= 0x10;  //Blinkled ein
                    delay(60);
                }else{
                    PORTC &= ~0x10; //Blinkled aus
                    delay(60);
                }
            }
        }


        //Tastatur abfragen
        //Taster ziehen nach Low
        PORTB = 62; //Spalte1 low: 2+4 + 56  (56=8+16+32 PullUps einschalten)
        asm ("nop");
        asm ("nop");
        if (!(PINB & 8)) tastencode=1;
        else
        if (!(PINB & 16)) tastencode=4;
        else
        if (!(PINB & 32)) tastencode=7;
        else{
            PORTB = 61; //Spalte2 low: 1+4 + 56  
            asm ("nop");
            asm ("nop");
            if (!(PINB & 8)) tastencode=2;
            else
            if (!(PINB & 16)) tastencode=5;
            else
            if (!(PINB & 32)) tastencode=8;
            else{
                PORTB = 59; //Spalte3 low: 1+2+ 56  (56=8+16+32)
                asm ("nop");
                asm ("nop");
                if (!(PINB & 8)) tastencode=3;
                else
                if (!(PINB & 16)) tastencode=6;
                else
                if (!(PINB & 32)) tastencode=9;
                else tastencode=0;
            }
        }
        PORTB = 63; //Spalten High, Pullups ein


        if (g.timeout_s > g.abschaltschwelle){
            tastencode = 9; //Abschalten
        }


        switch (tastencode){
        //Steckdosen:
        /*
         1= Receiver
         2= Monitor
         3= M3PO
         4= DVD
         5= WII
         6= VHS
         7= PS2
         8= Lampe
        */
        //1= Einschalten 0=Ausschalten x=don't care T=Toggle D=Ausschalten nach 8s
            case 1: g.pattern="1100000x"; break; //TV
            case 2: g.pattern="1010000x"; break; //MP3
            case 3: g.pattern="1101000x"; break; //DVD
            case 4: g.pattern="1100100x"; break; //Wii
            case 5: g.pattern="1100010x"; break; //VHS
            case 6: g.pattern="1100001x"; break; //PS2
            case 7: g.pattern="1000000x"; break; //Radio
            case 8: g.pattern="xxxxxxxT"; break; //Licht
            case 9: g.pattern="0000000D"; break; //Aus
        }

        if ((tastencode>0) && (g.tasteBisher != tastencode)){

       
            g.timeout_s=0;

            PORTC |= 0x10;  //Blinkled ein
            delay(60);
            PORTC &= ~0x10; //Blinkled aus
           
            PORTC = 32;    //Triac einschalten
            //warte200ms, bis Netzteil hochgefahren
            delay(200);

            //Relais setzen:
    
            //Einzuschaltende Relais in Zeile 1
            for (i=0; i<4; i++){
                if (     ('1'==g.pattern[i])
                     || (('T'==g.pattern[i]) && !g.toggleBisher[i] )
                  )
                {
                    PORTD=0x10;             //Zeilentreiber ein
                    PORTC=32+Relaiscode[i]; //Relais einschalten   
                    g.toggleBisher[i]=9;
                    delay(60); //Warte 60ms
                }
            }
            //Alle Spalten aus. 
            PORTC=32;
            //Warte 50ms bis Induktionsstrom abgebaut.
            nickerchen(); 
            delay(50); 

            //Einzuschaltende Relais in Zeile 3
            for (i=0; i<4; i++){
                if (    ('1'==g.pattern[i+4])
                    || (('T'==g.pattern[i+4]) && !g.toggleBisher[i+4] )
                   )
                {
                    PORTD=0x40;
                    //Relais einschalten   
                    PORTC=32+Relaiscode[i]; //Relais einschalten   
                    g.toggleBisher[i+4]=9;
                    delay(60); //Warte 60ms
                }
            }
            //Alle Spalten aus. 
            PORTC=32;
            //Warte 50ms bis Induktionsstrom abgebaut.
            nickerchen(); 
            delay(50); 


            //Auszuschaltende Relais in Zeile 2
            for (i=0; i<4; i++){
                if (     ('0'==g.pattern[i])
                     || (('T'==g.pattern[i]) && (1==g.toggleBisher[i]) )
                   )
                {
                    PORTD=0x20;
                    PORTC=32+Relaiscode[i]; //Relais ausschalten   
                    g.toggleBisher[i]=0;
                    delay(60); //Warte 60ms
                }
            }
            //Alle Spalten aus. 
            PORTC=32;
            //Warte 50ms bis Induktionsstrom abgebaut.
            nickerchen(); 
            delay(50); 

            //Auszuschaltende Relais in Zeile 4
            for (i=0; i<4; i++){
                if (     ('0'==g.pattern[i+4])
                     ||  ('D'==g.pattern[i+4])  
                     || (('T'==g.pattern[i+4]) && (1==g.toggleBisher[i+4]) )
                   )
                {
                    PORTD=0x80;
                    if ('D'==g.pattern[i+4]){
                      //Toggle ausschalten um 8s verzgern 
                      //(Fr Licht zum verlassen des Raumes)
                      for (n=0; n<8*5; n++){
                        delay(200);
                      }
                    }
                    PORTC=32+Relaiscode[i]; //Relais ausschalten   
                    g.toggleBisher[i+4]=0;
                    delay(60); //Warte 60ms
                }
            }
            //Damit beim Toggeln nach Einschalten nicht sofort wieder ausgeschaltet wird,
            //wird der "bisher" Wert auf 9 gesetzt und hier ggf. wieder auf 1
            //zurckgesetzt
            for (i=0; i<8; i++){
              g.toggleBisher[i] &=1;
            }

        }
        g.tasteBisher = tastencode;
        //Alle Zeilen, Spalten und Triac aus. 
        PORTC=0;
        PORTD=0;
        nickerchen(); 
        delay(40);    
    }
    return (0);
}
//EOF
